@@ -1,11 +1,15 @@
-import warnings
-from hashlib import md5
 import io
-import os
 import logging
+import os
+import warnings
+from distutils.version import LooseVersion
+from errno import ESPIPE
+from hashlib import sha256
+from glob import has_magic
 
+from .dircache import DirCache
 from .transaction import Transaction
-from .utils import read_block, tokenize, stringify_path
+from .utils import read_block, tokenize, stringify_path, other_paths
 
 logger = logging.getLogger("fsspec")
 
@@ -31,22 +35,24 @@ class _Cached(type):
     be made for a filesystem instance to be garbage collected.
     """
 
-    cachable = True
-    _extra_tokenize_attributes = ()
-
     def __init__(cls, *args, **kwargs):
         super().__init__(*args, **kwargs)
         # Note: we intentionally create a reference here, to avoid garbage
         # collecting instances when all other references are gone. To really
         # delete a FileSystem, the cache must be cleared.
         cls._cache = {}
+        cls._pid = os.getpid()
 
     def __call__(cls, *args, **kwargs):
         extra_tokens = tuple(
             getattr(cls, attr, None) for attr in cls._extra_tokenize_attributes
         )
         token = tokenize(cls, *args, *extra_tokens, **kwargs)
-        if cls.cachable and token in cls._cache:
+        skip = kwargs.pop("skip_instance_cache", False)
+        if os.getpid() != cls._pid:
+            cls._cache.clear()
+            cls._pid = os.getpid()
+        if not skip and cls.cachable and token in cls._cache:
             return cls._cache[token]
         else:
             obj = super().__call__(*args, **kwargs)
@@ -54,18 +60,26 @@ class _Cached(type):
             obj._fs_token_ = token
             obj.storage_args = args
             obj.storage_options = kwargs
+            if obj.async_impl:
+                from .asyn import mirror_sync_methods
 
-            if cls.cachable:
+                mirror_sync_methods(obj)
+
+            if cls.cachable and not skip:
                 cls._cache[token] = obj
             return obj
 
 
 try:  # optionally derive from pyarrow's FileSystem, if available
     import pyarrow as pa
-
-    up = pa.filesystem.DaskFileSystem
 except ImportError:
     up = object
+else:
+    # only derive from the legacy pyarrow's FileSystem for older pyarrow versions
+    if LooseVersion(pa.__version__) < LooseVersion("2.0"):
+        up = pa.filesystem.DaskFileSystem
+    else:
+        up = object
 
 
 class AbstractFileSystem(up, metaclass=_Cached):
@@ -81,6 +95,7 @@ class AbstractFileSystem(up, metaclass=_Cached):
     blocksize = 2 ** 22
     sep = "/"
     protocol = "abstract"
+    async_impl = False
     root_marker = ""  # For some FSs, may require leading '/' or other character
 
     #: Extra *class attributes* that should be considered when hashing.
@@ -97,9 +112,18 @@ class AbstractFileSystem(up, metaclass=_Cached):
 
         Subclasses should call this method.
 
-        Magic kwargs that affect functionality here:
-        add_docs: if True, will append docstrings from this spec to the
-            specific implementation
+        Parameters
+        ----------
+        use_listings_cache, listings_expiry_time, max_paths:
+            passed to ``DirCache``, if the implementation supports
+            directory listing caching. Pass use_listings_cache=False
+            to disable such caching.
+        skip_instance_cache: bool
+            If this is a cachable implementation, pass True here to force
+            creating a new instance even if a matching instance exists, and prevent
+            storing this instance.
+        asynchronous: bool
+        loop: asyncio-compatible IOLoop or None
         """
         if self._cached:
             # reusing instance, don't change
@@ -107,7 +131,8 @@ class AbstractFileSystem(up, metaclass=_Cached):
         self._cached = True
         self._intrans = False
         self._transaction = None
-        self.dircache = {}
+        self._invalidated_caches_in_transaction = []
+        self.dircache = DirCache(**storage_options)
 
         if storage_options.pop("add_docs", None):
             warnings.warn("add_docs is no longer supported.", FutureWarning)
@@ -130,25 +155,30 @@ class AbstractFileSystem(up, metaclass=_Cached):
     def __eq__(self, other):
         return isinstance(other, type(self)) and self._fs_token == other._fs_token
 
+    def __reduce__(self):
+        return make_instance, (type(self), self.storage_args, self.storage_options)
+
     @classmethod
     def _strip_protocol(cls, path):
-        """ Turn path from fully-qualified to file-system-specific
+        """Turn path from fully-qualified to file-system-specific
 
         May require FS-specific handling, e.g., for relative paths or links.
         """
+        if isinstance(path, list):
+            return [cls._strip_protocol(p) for p in path]
         path = stringify_path(path)
         protos = (cls.protocol,) if isinstance(cls.protocol, str) else cls.protocol
         for protocol in protos:
-            path = path.rstrip("/")
             if path.startswith(protocol + "://"):
                 path = path[len(protocol) + 3 :]
-            elif path.startswith(protocol + ":"):
-                path = path[len(protocol) + 1 :]
+            elif path.startswith(protocol + "::"):
+                path = path[len(protocol) + 2 :]
+        path = path.rstrip("/")
         # use of root_marker to make minimum required path, e.g., "/"
         return path or cls.root_marker
 
     @staticmethod
-    def _get_kwargs_from_urls(paths):
+    def _get_kwargs_from_urls(path):
         """If kwargs can be encoded in the paths, extract them here
 
         This should happen before instantiation of the class; incoming paths
@@ -162,7 +192,7 @@ class AbstractFileSystem(up, metaclass=_Cached):
 
     @classmethod
     def current(cls):
-        """ Return the most recently created FileSystem
+        """Return the most recently created FileSystem
 
         If no instance has been created, then create one with defaults
         """
@@ -192,6 +222,10 @@ class AbstractFileSystem(up, metaclass=_Cached):
         """Finish write transaction, non-context version"""
         self.transaction.complete()
         self._transaction = None
+        # The invalid cache must be cleared after the transcation is completed.
+        for path in self._invalidated_caches_in_transaction:
+            self.invalidate_cache(path)
+        self._invalidated_caches_in_transaction.clear()
 
     def invalidate_cache(self, path=None):
         """
@@ -203,7 +237,12 @@ class AbstractFileSystem(up, metaclass=_Cached):
             If None, clear all listings cached else listings at or under given
             path.
         """
-        pass  # not necessary to implement, may have no cache
+        # Not necessary to implement invalidation mechanism, may have no cache.
+        # But if have, you should call this method of parent class from your
+        # subclass to ensure expiring caches after transacations correctly.
+        # See the implementaion of FTPFileSystem in ftp.py
+        if self._intrans:
+            self._invalidated_caches_in_transaction.append(path)
 
     def mkdir(self, path, create_parents=True, **kwargs):
         """
@@ -253,6 +292,7 @@ class AbstractFileSystem(up, metaclass=_Cached):
         The specific keys, or perhaps a FileInfo class, or similar, is TBD,
         but must be consistent across implementations.
         Must include:
+
         - full path to the entry (without protocol)
         - size of the entry, in bytes. If the value cannot be determined, will
           be ``None``.
@@ -290,17 +330,21 @@ class AbstractFileSystem(up, metaclass=_Cached):
         but contains nothing), None if not in cache.
         """
         parent = self._parent(path)
-        if path in self.dircache:
+        try:
             return self.dircache[path]
-        elif parent in self.dircache:
+        except KeyError:
+            pass
+        try:
             files = [f for f in self.dircache[parent] if f["name"] == path]
             if len(files) == 0:
                 # parent dir was listed but did not contain this file
                 raise FileNotFoundError(path)
             return files
+        except KeyError:
+            pass
 
     def walk(self, path, maxdepth=None, **kwargs):
-        """ Return all files belows path
+        """Return all files belows path
 
         List all files, recursing into subdirectories; output is iterator-style,
         like ``os.walk()``. For a simple list of files, ``find()`` is available.
@@ -318,10 +362,11 @@ class AbstractFileSystem(up, metaclass=_Cached):
         kwargs: passed to ``ls``
         """
         path = self._strip_protocol(path)
-        full_dirs = []
-        dirs = []
-        files = []
+        full_dirs = {}
+        dirs = {}
+        files = {}
 
+        detail = kwargs.pop("detail", False)
         try:
             listing = self.ls(path, detail=True, **kwargs)
         except (FileNotFoundError, IOError):
@@ -330,26 +375,30 @@ class AbstractFileSystem(up, metaclass=_Cached):
         for info in listing:
             # each info name must be at least [path]/part , but here
             # we check also for names like [path]/part/
-            name = info["name"].rstrip("/")
-            if info["type"] == "directory" and name != path:
+            pathname = info["name"].rstrip("/")
+            name = pathname.rsplit("/", 1)[-1]
+            if info["type"] == "directory" and pathname != path:
                 # do not include "self" path
-                full_dirs.append(name)
-                dirs.append(name.rsplit("/", 1)[-1])
-            elif name == path:
+                full_dirs[pathname] = info
+                dirs[name] = info
+            elif pathname == path:
                 # file-like with same name as give path
-                files.append("")
+                files[""] = info
             else:
-                files.append(name.rsplit("/", 1)[-1])
-        yield path, dirs, files
+                files[name] = info
+
+        if detail:
+            yield path, dirs, files
+        else:
+            yield path, list(dirs), list(files)
+
+        if maxdepth is not None:
+            maxdepth -= 1
+            if maxdepth < 1:
+                return
 
         for d in full_dirs:
-            if maxdepth is None or maxdepth > 1:
-                for res in self.walk(
-                    d,
-                    maxdepth=(maxdepth - 1) if maxdepth is not None else None,
-                    **kwargs
-                ):
-                    yield res
+            yield from self.walk(d, maxdepth=maxdepth, detail=detail, **kwargs)
 
     def find(self, path, maxdepth=None, withdirs=False, **kwargs):
         """List all files below path.
@@ -367,18 +416,22 @@ class AbstractFileSystem(up, metaclass=_Cached):
         kwargs are passed to ``ls``.
         """
         # TODO: allow equivalent of -name parameter
-        out = set()
-        for path, dirs, files in self.walk(path, maxdepth, **kwargs):
+        path = self._strip_protocol(path)
+        out = dict()
+        detail = kwargs.pop("detail", False)
+        for path, dirs, files in self.walk(path, maxdepth, detail=True, **kwargs):
             if withdirs:
-                files += dirs
-            for name in files:
-                if name and name not in out:
-                    out.add("/".join([path.rstrip("/"), name]) if path else name)
+                files.update(dirs)
+            out.update({info["name"]: info for name, info in files.items()})
         if self.isfile(path) and path not in out:
             # walk works on directories, but find should also return [path]
             # when path happens to be a file
-            out.add(path)
-        return sorted(out)
+            out[path] = {}
+        names = sorted(out)
+        if not detail:
+            return names
+        else:
+            return {name: out[name] for name in names}
 
     def du(self, path, total=True, maxdepth=None, **kwargs):
         """Space used by files within a path
@@ -414,12 +467,15 @@ class AbstractFileSystem(up, metaclass=_Cached):
         the same as ``ls(path)``, returning only files.
 
         We support ``"**"``,
-        ``"?"`` and ``"[..]"``.
+        ``"?"`` and ``"[..]"``. We do not support ^ for pattern negation.
+
+        Search path names that contain embedded characters special to this
+        implementation of glob may not produce expected results;
+        e.g., 'foo/bar/*starredfilename*'.
 
         kwargs are passed to ``ls``.
         """
         import re
-        from glob import has_magic
 
         ends = path.endswith("/")
         path = self._strip_protocol(path)
@@ -429,23 +485,36 @@ class AbstractFileSystem(up, metaclass=_Cached):
 
         ind = min(indstar, indques, indbrace)
 
+        detail = kwargs.pop("detail", False)
+
         if not has_magic(path):
             root = path
             depth = 1
             if ends:
                 path += "/*"
             elif self.exists(path):
-                return [path]
+                if not detail:
+                    return [path]
+                else:
+                    return {path: self.info(path)}
             else:
-                return []  # glob of non-existent returns empty
+                if not detail:
+                    return []  # glob of non-existent returns empty
+                else:
+                    return {}
         elif "/" in path[:ind]:
             ind2 = path[:ind].rindex("/")
             root = path[: ind2 + 1]
-            depth = 20 if "**" in path else path[ind2 + 1 :].count("/") + 1
+            depth = None if "**" in path else path[ind2 + 1 :].count("/") + 1
         else:
             root = ""
-            depth = 20 if "**" in path else 1
-        allpaths = self.find(root, maxdepth=depth, withdirs=True, **kwargs)
+            depth = None if "**" in path else 1
+
+        allpaths = self.find(root, maxdepth=depth, withdirs=True, detail=True, **kwargs)
+        # Escape characters special to python regex, leaving our supported
+        # special characters in place.
+        # See https://www.gnu.org/software/bash/manual/html_node/Pattern-Matching.html
+        # for shell globbing details.
         pattern = (
             "^"
             + (
@@ -456,6 +525,10 @@ class AbstractFileSystem(up, metaclass=_Cached):
                 .replace("(", r"\(")
                 .replace(")", r"\)")
                 .replace("|", r"\|")
+                .replace("^", r"\^")
+                .replace("$", r"\$")
+                .replace("{", r"\{")
+                .replace("}", r"\}")
                 .rstrip("/")
                 .replace("?", ".")
             )
@@ -464,8 +537,15 @@ class AbstractFileSystem(up, metaclass=_Cached):
         pattern = re.sub("[*]{2}", "=PLACEHOLDER=", pattern)
         pattern = re.sub("[*]", "[^/]*", pattern)
         pattern = re.compile(pattern.replace("=PLACEHOLDER=", ".*"))
-        out = {p for p in allpaths if pattern.match(p.replace("//", "/").rstrip("/"))}
-        return list(sorted(out))
+        out = {
+            p: allpaths[p]
+            for p in sorted(allpaths)
+            if pattern.match(p.replace("//", "/").rstrip("/"))
+        }
+        if detail:
+            return out
+        else:
+            return list(out)
 
     def exists(self, path):
         """Is there a file at the given path"""
@@ -531,7 +611,7 @@ class AbstractFileSystem(up, metaclass=_Cached):
         """Is this entry directory-like?"""
         try:
             return self.info(path)["type"] == "directory"
-        except FileNotFoundError:
+        except IOError:
             return False
 
     def isfile(self, path):
@@ -541,30 +621,73 @@ class AbstractFileSystem(up, metaclass=_Cached):
         except:  # noqa: E722
             return False
 
-    def cat(self, path):
+    def cat_file(self, path):
         """ Get the content of a file """
         return self.open(path, "rb").read()
 
-    def get(self, rpath, lpath, recursive=False, **kwargs):
-        """Copy file to local.
+    def pipe_file(self, path, value, **kwargs):
+        """Set the bytes of given file"""
+        with self.open(path, "wb") as f:
+            f.write(value)
 
-        Possible extension: maybe should be able to copy to any file-system
-        (streaming through local).
+    def pipe(self, path, value=None, **kwargs):
+        """Put value into path
+
+        (counterpart to ``cat``)
+        Parameters
+        ----------
+        path: string or dict(str, bytes)
+            If a string, a single remote location to put ``value`` bytes; if a dict,
+            a mapping of {path: bytesvalue}.
+        value: bytes, optional
+            If using a single path, these are the bytes to put there. Ignored if
+            ``path`` is a dict
         """
-        rpath = self._strip_protocol(rpath)
-        if recursive:
-            rpaths = self.find(rpath)
-            lpaths = [
-                os.path.join(lpath, path[len(rpath) :].lstrip("/")) for path in rpaths
-            ]
-            for lpath in lpaths:
-                dirname = os.path.dirname(lpath)
-                if not os.path.isdir(dirname):
-                    os.makedirs(dirname)
+        if isinstance(path, str):
+            self.pipe_file(self._strip_protocol(path), value, **kwargs)
+        elif isinstance(path, dict):
+            for k, v in path.items():
+                self.pipe_file(self._strip_protocol(k), v, **kwargs)
+        else:
+            raise ValueError("path must be str or dict")
+
+    def cat(self, path, recursive=False, on_error="raise", **kwargs):
+        """Fetch (potentially multiple) paths' contents
+
+        Returns a dict of {path: contents} if there are multiple paths
+        or the path has been otherwise expanded
+
+        on_error : "raise", "omit", "return"
+            If raise, an underlying exception will be raised (converted to KeyError
+            if the type is in self.missing_exceptions); if omit, keys with exception
+            will simply not be included in the output; if "return", all keys are
+            included in the output, but the value will be bytes or an exception
+            instance.
+        """
+        paths = self.expand_path(path, recursive=recursive)
+        if (
+            len(paths) > 1
+            or isinstance(path, list)
+            or paths[0] != self._strip_protocol(path)
+        ):
+            out = {}
+            for path in paths:
+                try:
+                    out[path] = self.cat_file(path, **kwargs)
+                except Exception as e:
+                    if on_error == "raise":
+                        raise
+                    if on_error == "return":
+                        out[path] = e
+            return out
+        else:
+            return self.cat_file(paths[0])
+
+    def get_file(self, rpath, lpath, **kwargs):
+        """Copy single remote file to local"""
+        if self.isdir(rpath):
+            os.makedirs(lpath, exist_ok=True)
         else:
-            rpaths = [rpath]
-            lpaths = [lpath]
-        for lpath, rpath in zip(lpaths, rpaths):
             with self.open(rpath, "rb", **kwargs) as f1:
                 with open(lpath, "wb") as f2:
                     data = True
@@ -572,33 +695,63 @@ class AbstractFileSystem(up, metaclass=_Cached):
                         data = f1.read(self.blocksize)
                         f2.write(data)
 
-    def put(self, lpath, rpath, recursive=False, **kwargs):
-        """ Upload file from local """
-        if recursive:
-            lpaths = []
-            for dirname, subdirlist, filelist in os.walk(lpath):
-                lpaths += [os.path.join(dirname, filename) for filename in filelist]
-            rootdir = os.path.basename(lpath.rstrip("/"))
-            if self.exists(rpath):
-                # copy lpath inside rpath directory
-                rpath2 = os.path.join(rpath, rootdir)
-            else:
-                # copy lpath as rpath directory
-                rpath2 = rpath
-            rpaths = [
-                os.path.join(rpath2, path[len(lpath) :].lstrip("/")) for path in lpaths
-            ]
-        else:
-            lpaths = [lpath]
-            rpaths = [rpath]
+    def get(self, rpath, lpath, recursive=False, **kwargs):
+        """Copy file(s) to local.
+
+        Copies a specific file or tree of files (if recursive=True). If lpath
+        ends with a "/", it will be assumed to be a directory, and target files
+        will go within. Can submit a list of paths, which may be glob-patterns
+        and will be expanded.
+
+        Calls get_file for each source.
+        """
+        from .implementations.local import make_path_posix
+
+        if isinstance(lpath, str):
+            lpath = make_path_posix(lpath)
+        rpaths = self.expand_path(rpath, recursive=recursive)
+        lpaths = other_paths(rpaths, lpath)
         for lpath, rpath in zip(lpaths, rpaths):
+            self.get_file(rpath, lpath, **kwargs)
+
+    def put_file(self, lpath, rpath, **kwargs):
+        """Copy single file to remote"""
+        if os.path.isdir(lpath):
+            self.makedirs(rpath, exist_ok=True)
+        else:
             with open(lpath, "rb") as f1:
+                self.mkdirs(os.path.dirname(rpath), exist_ok=True)
                 with self.open(rpath, "wb", **kwargs) as f2:
                     data = True
                     while data:
                         data = f1.read(self.blocksize)
                         f2.write(data)
 
+    def put(self, lpath, rpath, recursive=False, **kwargs):
+        """Copy file(s) from local.
+
+        Copies a specific file or tree of files (if recursive=True). If rpath
+        ends with a "/", it will be assumed to be a directory, and target files
+        will go within.
+
+        Calls put_file for each source.
+        """
+        from .implementations.local import make_path_posix, LocalFileSystem
+
+        rpath = (
+            self._strip_protocol(rpath)
+            if isinstance(rpath, str)
+            else [self._strip_protocol(p) for p in rpath]
+        )
+        if isinstance(lpath, str):
+            lpath = make_path_posix(lpath)
+        fs = LocalFileSystem()
+        lpaths = fs.expand_path(lpath, recursive=recursive)
+        rpaths = other_paths(lpaths, rpath)
+
+        for lpath, rpath in zip(lpaths, rpaths):
+            self.put_file(lpath, rpath, **kwargs)
+
     def head(self, path, size=1024):
         """ Get the first ``size`` bytes from file """
         with self.open(path, "rb") as f:
@@ -610,17 +763,52 @@ class AbstractFileSystem(up, metaclass=_Cached):
             f.seek(max(-size, -f.size), 2)
             return f.read()
 
-    def copy(self, path1, path2, **kwargs):
-        """ Copy within two locations in the filesystem"""
+    def cp_file(self, path1, path2, **kwargs):
         raise NotImplementedError
 
-    def mv(self, path1, path2, **kwargs):
-        """ Move file from one location to another """
-        self.copy(path1, path2, **kwargs)
-        self.rm(path1, recursive=False)
+    def copy(self, path1, path2, recursive=False, **kwargs):
+        """ Copy within two locations in the filesystem"""
+        paths = self.expand_path(path1, recursive=recursive)
+        path2 = other_paths(paths, path2)
+        for p1, p2 in zip(paths, path2):
+            self.cp_file(p1, p2, **kwargs)
+
+    def expand_path(self, path, recursive=False, maxdepth=None):
+        """Turn one or more globs or directories into a list of all matching files"""
+        if isinstance(path, str):
+            out = self.expand_path([path], recursive, maxdepth)
+        else:
+            out = set()
+            path = [self._strip_protocol(p) for p in path]
+            for p in path:
+                if has_magic(p):
+                    bit = set(self.glob(p))
+                    out |= bit
+                    if recursive:
+                        out += self.expand_path(p)
+                    continue
+                elif recursive:
+                    rec = set(self.find(p, withdirs=True))
+                    out |= rec
+                if p not in out and (recursive is False or self.exists(p)):
+                    # should only check once, for the root
+                    out.add(p)
+        if not out:
+            raise FileNotFoundError(path)
+        return list(sorted(out))
+
+    def mv(self, path1, path2, recursive=False, maxdepth=None, **kwargs):
+        """ Move file(s) from one location to another """
+        self.copy(path1, path2, recursive=recursive, maxdepth=maxdepth)
+        self.rm(path1, recursive=recursive)
+
+    def rm_file(self, path):
+        """Delete a file"""
+        self._rm(path)
 
     def _rm(self, path):
-        """Delete a file"""
+        """Delete one file"""
+        # this is the old name for the method, prefer rm_file
         raise NotImplementedError
 
     def rm(self, path, recursive=False, maxdepth=None):
@@ -638,25 +826,16 @@ class AbstractFileSystem(up, metaclass=_Cached):
             If None, there will be no limit and infinite recursion may be
             possible.
         """
-        # prefer some bulk method, if possible
-        if not isinstance(path, list):
-            path = [path]
-        for p in path:
-            if recursive:
-                out = self.walk(p, maxdepth=maxdepth)
-                for pa_, _, files in reversed(list(out)):
-                    for name in files:
-                        fn = "/".join([pa_, name]) if pa_ else name
-                        self.rm(fn)
-                    self.rmdir(pa_)
-            else:
-                self._rm(p)
+        path = self.expand_path(path, recursive=recursive, maxdepth=maxdepth)
+        for p in reversed(path):
+            self.rm_file(p)
 
     @classmethod
     def _parent(cls, path):
         path = cls._strip_protocol(path.rstrip("/"))
         if "/" in path:
-            return cls.root_marker + path.rsplit("/", 1)[0]
+            parent = path.rsplit("/", 1)[0].lstrip(cls.root_marker)
+            return cls.root_marker + parent
         else:
             return cls.root_marker
 
@@ -728,7 +907,7 @@ class AbstractFileSystem(up, metaclass=_Cached):
             return f
 
     def touch(self, path, truncate=True, **kwargs):
-        """ Create empty file, or update timestamp
+        """Create empty file, or update timestamp
 
         Parameters
         ----------
@@ -746,10 +925,10 @@ class AbstractFileSystem(up, metaclass=_Cached):
 
     def ukey(self, path):
         """Hash of file properties, to tell if it has changed"""
-        return md5(str(self.info(path)).encode()).hexdigest()
+        return sha256(str(self.info(path)).encode()).hexdigest()
 
     def read_block(self, fn, offset, length, delimiter=None):
-        """ Read a block of bytes from
+        """Read a block of bytes from
 
         Starting at ``offset`` of the file, read ``length`` bytes.  If
         ``delimiter`` is set then we ensure that the read starts and stops at
@@ -793,8 +972,58 @@ class AbstractFileSystem(up, metaclass=_Cached):
                 length = size - offset
             return read_block(f, offset, length, delimiter)
 
-    def __reduce__(self):
-        return make_instance, (type(self), self.storage_args, self.storage_options)
+    def to_json(self):
+        """
+        JSON representation of this filesystem instance
+
+        Returns
+        -------
+        str: JSON structure with keys cls (the python location of this class),
+            protocol (text name of this class's protocol, first one in case of
+            multiple), args (positional args, usually empty), and all other
+            kwargs as their own keys.
+        """
+        import json
+
+        cls = type(self)
+        cls = ".".join((cls.__module__, cls.__name__))
+        proto = (
+            self.protocol[0]
+            if isinstance(self.protocol, (tuple, list))
+            else self.protocol
+        )
+        return json.dumps(
+            dict(
+                **{"cls": cls, "protocol": proto, "args": self.storage_args},
+                **self.storage_options
+            )
+        )
+
+    @staticmethod
+    def from_json(blob):
+        """
+        Recreate a filesystem instance from JSON representation
+
+        See ``.to_json()`` for the expected structure of the input
+
+        Parameters
+        ----------
+        blob: str
+
+        Returns
+        -------
+        file system instance, not necessarily of this particular class.
+        """
+        from .registry import _import_class, get_filesystem_class
+        import json
+
+        dic = json.loads(blob)
+        protocol = dic.pop("protocol")
+        try:
+            cls = _import_class(dic.pop("cls"))
+        except (ImportError, ValueError, RuntimeError, KeyError):
+            cls = get_filesystem_class(protocol)
+        return cls(*dic.pop("args", ()), **dic)
 
     def _get_pyarrow_filesystem(self):
         """
@@ -828,6 +1057,14 @@ class AbstractFileSystem(up, metaclass=_Cached):
         """
         cls._cache.clear()
 
+    def created(self, path):
+        """Return the created timestamp of a file as a datetime.datetime"""
+        raise NotImplementedError
+
+    def modified(self, path):
+        """Return the modified timestamp of a file as a datetime.datetime"""
+        raise NotImplementedError
+
     # ------------------------------------------------------------------------
     # Aliases
 
@@ -875,6 +1112,37 @@ class AbstractFileSystem(up, metaclass=_Cached):
         """Alias of :ref:`FilesystemSpec.get`."""
         return self.get(rpath, lpath, recursive=recursive, **kwargs)
 
+    def sign(self, path, expiration=100, **kwargs):
+        """Create a signed URL representing the given path
+
+        Some implementations allow temporary URLs to be generated, as a
+        way of delegating credentials.
+
+        Parameters
+        ----------
+        path : str
+             The path on the filesystem
+        expiration : int
+            Number of seconds to enable the URL for (if supported)
+
+        Returns
+        -------
+        URL : str
+            The signed URL
+
+        Raises
+        ------
+        NotImplementedError : if method is not implemented for a fileystem
+        """
+        raise NotImplementedError("Sign is not implemented for this filesystem")
+
+    def _isfilestore(self):
+        # Originally inherited from pyarrow DaskFileSystem. Keeping this
+        # here for backwards compatibility as long as pyarrow uses its
+        # legacy ffspec-compatible filesystems and thus accepts fsspec
+        # filesystems as well
+        return False
+
 
 class AbstractBufferedFile(io.IOBase):
     """Convenient class to derive from to provide buffering
@@ -967,7 +1235,8 @@ class AbstractBufferedFile(io.IOBase):
     @property
     def closed(self):
         # get around this attr being read-only in IOBase
-        return self._closed
+        # use getattr here, since this can be called during del
+        return getattr(self, "_closed", True)
 
     @closed.setter
     def closed(self, c):
@@ -1001,7 +1270,7 @@ class AbstractBufferedFile(io.IOBase):
         return self.loc
 
     def seek(self, loc, whence=0):
-        """ Set current file location
+        """Set current file location
 
         Parameters
         ----------
@@ -1012,7 +1281,7 @@ class AbstractBufferedFile(io.IOBase):
         """
         loc = int(loc)
         if not self.mode == "rb":
-            raise ValueError("Seek only available in read mode")
+            raise OSError(ESPIPE, "Seek only available in read mode")
         if whence == 0:
             nloc = loc
         elif whence == 1:
@@ -1089,7 +1358,7 @@ class AbstractBufferedFile(io.IOBase):
             self.buffer = io.BytesIO()
 
     def _upload_chunk(self, final=False):
-        """ Write one part of a multi-block file upload
+        """Write one part of a multi-block file upload
 
         Parameters
         ==========
@@ -1097,7 +1366,7 @@ class AbstractBufferedFile(io.IOBase):
             This is the last block, so should complete file, if
             self.autocommit is True.
         """
-        # may not yet have been initialized, may neet to call _initialize_upload
+        # may not yet have been initialized, may need to call _initialize_upload
 
     def _initiate_upload(self):
         """ Create remote file/upload """
@@ -1136,8 +1405,9 @@ class AbstractBufferedFile(io.IOBase):
 
         https://docs.python.org/3/library/io.html#io.RawIOBase.readinto
         """
-        data = self.read(len(b))
-        b[: len(data)] = data
+        out = memoryview(b).cast("B")
+        data = self.read(out.nbytes)
+        out[: len(data)] = data
         return len(data)
 
     def readuntil(self, char=b"\n", blocks=None):
@@ -1200,7 +1470,7 @@ class AbstractBufferedFile(io.IOBase):
         return self.readinto(b)
 
     def close(self):
-        """ Close file
+        """Close file
 
         Finalizes writes, discards cache
         """
@@ -1231,7 +1501,8 @@ class AbstractBufferedFile(io.IOBase):
         return self.mode in {"wb", "ab"} and not self.closed
 
     def __del__(self):
-        self.close()
+        if not self.closed:
+            self.close()
 
     def __str__(self):
         return "<File-like object %s, %s>" % (type(self.fs).__name__, self.path)
