@@ -5,7 +5,11 @@ from multiprocessing import Process
 
 import pytest
 
-pytest.importorskip("fuse")  # noqa: E402
+try:
+    pytest.importorskip("fuse")  # noqa: E402
+except OSError:
+    # can succeed in importing fuse, but fail to load so
+    pytest.importorskip("nonexistent")  # noqa: E402
 
 from fsspec.fuse import run
 from fsspec.implementations.memory import MemoryFileSystem
@@ -17,7 +21,7 @@ def host_fuse(mountdir):
     run(fs, "/mounted/", mountdir)
 
 
-def test_basic(tmpdir):
+def test_basic(tmpdir, capfd):
     mountdir = str(tmpdir.mkdir("mount"))
 
     fuse_process = Process(target=host_fuse, args=(str(mountdir),))
@@ -53,6 +57,10 @@ def test_basic(tmpdir):
         with pytest.raises(OSError):
             os.rmdir(fn)
 
+        captured = capfd.readouterr()
+        assert "Traceback" not in captured.out
+        assert "Traceback" not in captured.err
+
         os.rmdir(fn + "/inner")
         os.rmdir(fn)
     finally:
